@@ -5,6 +5,7 @@ import (
 	"crypto/sha256"
 	"crypto/tls"
 	"encoding/base64"
+	"encoding/hex"
 	"encoding/json"
 	"errors"
 	"flag"
@@ -13,12 +14,14 @@ import (
 	"net/http"
 	"os"
 	"os/signal"
+	"sort"
 	"strings"
 	"sync"
 	"syscall"
 	"time"
 
 	rtr "github.com/bgp/stayrtr/lib"
+	"github.com/bgp/stayrtr/ossec"
 	"github.com/bgp/stayrtr/prefixfile"
 	"github.com/bgp/stayrtr/utils"
 	"github.com/prometheus/client_golang/prometheus"
@@ -50,8 +53,7 @@ var (
 
 	ExportPath = flag.String("export.path", "/rpki.json", "Export path")
 
-	RTRVersion = flag.Int("protocol", 1, "RTR protocol version")
-	SessionID  = flag.Int("rtr.sessionid", -1, "Set session ID (if < 0: will be randomized)")
+	RTRVersion = flag.Int("protocol", 2, "RTR protocol version. Version 2 is draft-ietf-sidrops-8210bis-10")
 	RefreshRTR = flag.Int("rtr.refresh", 3600, "Refresh interval")
 	RetryRTR   = flag.Int("rtr.retry", 600, "Retry interval")
 	ExpireRTR  = flag.Int("rtr.expire", 7200, "Expire interval")
@@ -99,6 +101,13 @@ var (
 		},
 		[]string{"ip_version", "filtered", "path"},
 	)
+	NumberOfBRKs = prometheus.NewGaugeVec(
+		prometheus.GaugeOpts{
+			Name: "rpki_bgpsec",
+			Help: "Number of BGPsec keys.",
+		},
+		[]string{},
+	)
 	LastRefresh = prometheus.NewGaugeVec(
 		prometheus.GaugeOpts{
 			Name: "rpki_refresh",
@@ -138,21 +147,13 @@ var (
 	protoverToLib = map[int]uint8{
 		0: rtr.PROTOCOL_VERSION_0,
 		1: rtr.PROTOCOL_VERSION_1,
-	}
-	authToId = map[string]int{
-		"none":     METHOD_NONE,
-		"password": METHOD_PASSWORD,
-		//"key":   METHOD_KEY,
-	}
-	serialToId = map[string]int{
-		"disable": USE_SERIAL_DISABLE,
-		"startup": USE_SERIAL_START,
-		"full":    USE_SERIAL_FULL,
+		2: rtr.PROTOCOL_VERSION_2,
 	}
 )
 
 func initMetrics() {
 	prometheus.MustRegister(NumberOfVRPs)
+	prometheus.MustRegister(NumberOfBRKs)
 	prometheus.MustRegister(LastChange)
 	prometheus.MustRegister(LastRefresh)
 	prometheus.MustRegister(RefreshStatusCode)
@@ -196,8 +197,11 @@ func isValidPrefixLength(prefix *net.IPNet, maxLength uint8) bool {
 // 1 - The prefix is a valid prefix
 // 2 - The ASN is a valid ASN
 // 3 - The MaxLength is valid
-// Will return a deduped slice, as well as total VRPs, IPv4 VRPs, and IPv6 VRPs
-func processData(vrplistjson []prefixfile.VRPJson) ([]rtr.VRP, int, int, int) {
+// Will return a deduped slice, as well as total VRPs, IPv4 VRPs, IPv6 VRPs, BGPsec Keys and ASPA records
+func processData(vrplistjson []prefixfile.VRPJson,
+	brklistjson []prefixfile.BgpSecKeyJson,
+	aspajson *prefixfile.ProviderAuthorizationsJson) /*Export*/ ([]rtr.VRP, []rtr.BgpsecKey, []rtr.VAP, int, int, int) {
+	//
 	filterDuplicates := make(map[string]bool)
 
 	// It may be tempting to change this to a simple time.Since() but that will
@@ -207,6 +211,8 @@ func processData(vrplistjson []prefixfile.VRPJson) ([]rtr.VRP, int, int, int) {
 	NowUnix := time.Now().UTC().Unix()
 
 	var vrplist []rtr.VRP
+	var brklist = make([]rtr.BgpsecKey, 0)
+	var aspalist = make([]rtr.VAP, 0)
 	var countv4 int
 	var countv6 int
 
@@ -254,7 +260,57 @@ func processData(vrplistjson []prefixfile.VRPJson) ([]rtr.VRP, int, int, int) {
 		}
 		vrplist = append(vrplist, vrp)
 	}
-	return vrplist, countv4 + countv6, countv4, countv6
+
+	for _, v := range brklistjson {
+		if v.Expires != nil {
+			// Prevent stale VRPs from being considered
+			// https://github.com/bgp/stayrtr/issues/15
+			if int(NowUnix) > int(*v.Expires) {
+				continue
+			}
+		}
+
+		SKIBytes, err := hex.DecodeString(v.Ski)
+		if err != nil {
+			continue
+		}
+
+		brklist = append(brklist, rtr.BgpsecKey{
+			ASN:    v.Asn,
+			Pubkey: v.Pubkey,
+			Ski:    SKIBytes,
+		})
+	}
+
+	if aspajson != nil {
+		aspalist = handleASPAList(aspajson.IPv4, NowUnix, aspalist, rtr.AFI_IPv4)
+		aspalist = handleASPAList(aspajson.IPv6, NowUnix, aspalist, rtr.AFI_IPv6)
+	}
+
+	return vrplist, brklist, aspalist, countv4 + countv6, countv4, countv6
+}
+
+func handleASPAList(list []prefixfile.ASPAJson, NowUnix int64, aspalist []rtr.VAP, AFI uint8) []rtr.VAP {
+	for _, v := range list {
+		if v.Expires != nil {
+			if int(NowUnix) > int(*v.Expires) {
+				continue
+			}
+		}
+
+		// Ensure that these are sorted, otherwise they
+		// don't has right.
+		sort.Slice(v.Providers, func(i, j int) bool {
+			return v.Providers[i] < v.Providers[j]
+		})
+
+		aspalist = append(aspalist, rtr.VAP{
+			AFI:         AFI,
+			CustomerASN: v.CustomerAsid,
+			Providers:   v.Providers,
+		})
+	}
+	return aspalist
 }
 
 type IdenticalFile struct {
@@ -275,8 +331,22 @@ func (s *state) updateFromNewState() error {
 	if vrpsjson == nil {
 		return nil
 	}
+	bgpsecjson := s.lastdata.BgpSecKeys
+	if bgpsecjson == nil {
+		bgpsecjson = make([]prefixfile.BgpSecKeyJson, 0)
+	}
+	aspajson := s.lastdata.ASPA
+	if aspajson == nil {
+		aspajson = &prefixfile.ProviderAuthorizationsJson{
+			IPv4: make([]prefixfile.ASPAJson, 0),
+			IPv6: make([]prefixfile.ASPAJson, 0),
+		}
+	}
 
 	buildtime, err := time.Parse(time.RFC3339, s.lastdata.Metadata.Buildtime)
+	if s.lastdata.Metadata.GeneratedUnix != nil {
+		buildtime, err = time.Unix(int64(*s.lastdata.Metadata.GeneratedUnix), 0), nil
+	}
 	if s.checktime {
 		if err != nil {
 			return err
@@ -289,16 +359,13 @@ func (s *state) updateFromNewState() error {
 	}
 
 	if s.slurm != nil {
-		kept, removed := s.slurm.FilterOnVRPs(vrpsjson)
-		asserted := s.slurm.AssertVRPs()
-		log.Infof("Slurm filtering: %v kept, %v removed, %v asserted", len(kept), len(removed), len(asserted))
-		vrpsjson = append(kept, asserted...)
+		vrpsjson, aspajson.IPv4, aspajson.IPv6, bgpsecjson = s.slurm.FilterAssert(vrpsjson, aspajson.IPv4, aspajson.IPv6, bgpsecjson, log.StandardLogger())
 	}
 
-	vrps, count, countv4, countv6 := processData(vrpsjson)
+	vrps, brks, vaps, count, countv4, countv6 := processData(vrpsjson, bgpsecjson, aspajson)
 
 	log.Infof("New update (%v uniques, %v total prefixes).", len(vrps), count)
-	return s.applyUpdateFromNewState(vrps, sessid, vrpsjson, countv4, countv6)
+	return s.applyUpdateFromNewState(vrps, brks, vaps, sessid, vrpsjson, bgpsecjson, aspajson, countv4, countv6)
 }
 
 // Update the state based on the currently loaded files
@@ -309,8 +376,22 @@ func (s *state) reloadFromCurrentState() error {
 	if vrpsjson == nil {
 		return nil
 	}
+	bgpsecjson := s.lastdata.BgpSecKeys
+	if bgpsecjson == nil {
+		bgpsecjson = make([]prefixfile.BgpSecKeyJson, 0)
+	}
+	aspajson := s.lastdata.ASPA
+	if aspajson == nil {
+		aspajson = &prefixfile.ProviderAuthorizationsJson{
+			IPv4: make([]prefixfile.ASPAJson, 0),
+			IPv6: make([]prefixfile.ASPAJson, 0),
+		}
+	}
 
 	buildtime, err := time.Parse(time.RFC3339, s.lastdata.Metadata.Buildtime)
+	if s.lastdata.Metadata.GeneratedUnix != nil {
+		buildtime, err = time.Unix(int64(*s.lastdata.Metadata.GeneratedUnix), 0), nil
+	}
 	if s.checktime {
 		if err != nil {
 			return err
@@ -323,22 +404,33 @@ func (s *state) reloadFromCurrentState() error {
 	}
 
 	if s.slurm != nil {
-		kept, removed := s.slurm.FilterOnVRPs(vrpsjson)
-		asserted := s.slurm.AssertVRPs()
-		log.Infof("Slurm filtering: %v kept, %v removed, %v asserted", len(kept), len(removed), len(asserted))
-		vrpsjson = append(kept, asserted...)
+		vrpsjson, aspajson.IPv4, aspajson.IPv6, bgpsecjson = s.slurm.FilterAssert(vrpsjson, aspajson.IPv4, aspajson.IPv6, bgpsecjson, log.StandardLogger())
 	}
 
-	vrps, count, countv4, countv6 := processData(vrpsjson)
+	vrps, brks, vaps, count, countv4, countv6 := processData(vrpsjson, bgpsecjson, aspajson)
 	if s.server.CountVRPs() != count {
 		log.Infof("New update to old state (%v uniques, %v total prefixes). (old %v - new %v)", len(vrps), count, s.server.CountVRPs(), count)
-		return s.applyUpdateFromNewState(vrps, sessid, vrpsjson, countv4, countv6)
+		return s.applyUpdateFromNewState(vrps, brks, vaps, sessid, vrpsjson, bgpsecjson, aspajson, countv4, countv6)
 	}
 	return nil
 }
 
-func (s *state) applyUpdateFromNewState(vrps []rtr.VRP, sessid uint16, vrpsjson []prefixfile.VRPJson, countv4 int, countv6 int) error {
-	s.server.AddVRPs(vrps)
+func (s *state) applyUpdateFromNewState(vrps []rtr.VRP, brks []rtr.BgpsecKey, vaps []rtr.VAP,
+	sessid uint16,
+	vrpsjson []prefixfile.VRPJson, brksjson []prefixfile.BgpSecKeyJson, aspajson *prefixfile.ProviderAuthorizationsJson,
+	countv4 int, countv6 int) error {
+
+	SDs := make([]rtr.SendableData, 0)
+	for _, v := range vrps {
+		SDs = append(SDs, v.Copy())
+	}
+	for _, v := range brks {
+		SDs = append(SDs, v.Copy())
+	}
+	for _, v := range vaps {
+		SDs = append(SDs, v.Copy())
+	}
+	s.server.AddData(SDs)
 
 	serial, _ := s.server.GetCurrentSerial(sessid)
 	log.Infof("Updated added, new serial %v", serial)
@@ -353,7 +445,9 @@ func (s *state) applyUpdateFromNewState(vrps []rtr.VRP, sessid uint16, vrpsjson
 			Counts:    len(vrpsjson),
 			Buildtime: s.lastdata.Metadata.Buildtime,
 		},
-		Data: vrpsjson,
+		Data:       vrpsjson,
+		BgpSecKeys: brksjson,
+		ASPA:       aspajson,
 	}
 
 	s.lockJson.Unlock()
@@ -368,7 +462,7 @@ func (s *state) applyUpdateFromNewState(vrps []rtr.VRP, sessid uint16, vrpsjson
 				countv6_dup++
 			}
 		}
-		s.metricsEvent.UpdateMetrics(countv4, countv6, countv4_dup, countv6_dup, s.lastchange, s.lastts, *CacheBin)
+		s.metricsEvent.UpdateMetrics(countv4, countv6, countv4_dup, countv6_dup, s.lastchange, s.lastts, *CacheBin, len(brks))
 	}
 
 	return nil
@@ -513,7 +607,7 @@ func (s *state) routineUpdate(file string, interval int, slurmFile string) {
 					// to avoid routing on stale data
 					buildTime := s.exported.Metadata.GetBuildTime()
 					if !buildTime.IsZero() && time.Since(buildTime) > time.Hour*24 {
-						s.server.AddVRPs([]rtr.VRP{}) // empty the store of VRP by giving it a empty VRP array, triggering a emptying
+						s.server.AddData([]rtr.SendableData{}) // empty the store of sendable stuff, triggering a emptying of the RTR server
 					}
 				}
 				log.Errorf("Error updating from new state: %v", err)
@@ -541,7 +635,6 @@ type state struct {
 	lastchange time.Time
 	lastts     time.Time
 	sendNotifs bool
-	useSerial  int
 
 	fetchConfig *utils.FetchConfig
 
@@ -578,15 +671,22 @@ func (m *metricsEvent) HandlePDU(c *rtr.Client, pdu rtr.PDU) {
 				"_", -1))).Inc()
 }
 
-func (m *metricsEvent) UpdateMetrics(numIPv4 int, numIPv6 int, numIPv4filtered int, numIPv6filtered int, changed time.Time, refreshed time.Time, file string) {
+func (m *metricsEvent) UpdateMetrics(numIPv4 int, numIPv6 int, numIPv4filtered int, numIPv6filtered int, changed time.Time, refreshed time.Time, file string, brkCount int) {
 	NumberOfVRPs.WithLabelValues("ipv4", "filtered", file).Set(float64(numIPv4filtered))
 	NumberOfVRPs.WithLabelValues("ipv4", "unfiltered", file).Set(float64(numIPv4))
 	NumberOfVRPs.WithLabelValues("ipv6", "filtered", file).Set(float64(numIPv6filtered))
 	NumberOfVRPs.WithLabelValues("ipv6", "unfiltered", file).Set(float64(numIPv6))
+	NumberOfBRKs.WithLabelValues().Set(float64(brkCount))
 	LastChange.WithLabelValues(file).Set(float64(changed.UnixNano() / 1e9))
 }
 
 func main() {
+	err := ossec.PledgePromises("dns inet rpath stdio tty")
+	if err != nil {
+		fmt.Fprintf(os.Stderr, "pledge failed: %v\n", err)
+		os.Exit(1)
+	}
+
 	if err := run(); err != nil {
 		fmt.Fprintf(os.Stderr, "%s\n", err)
 		os.Exit(1)
@@ -613,7 +713,6 @@ func run() error {
 
 	sc := rtr.ServerConfiguration{
 		ProtocolVersion: protoverToLib[*RTRVersion],
-		SessId:          *SessionID,
 		KeepDifference:  3,
 		Log:             log.StandardLogger(),
 		LogVerbose:      *LogVerbose,
@@ -632,7 +731,7 @@ func run() error {
 	}
 
 	server := rtr.NewServer(sc, me, deh)
-	deh.SetVRPManager(server)
+	deh.SetSDManager(server)
 
 	s := state{
 		server:       server,
